Brick = class("Brick", Sprite)

function Brick:initialize(x, y)
	Sprite.initialize(self, "brick_spritesheet", rects.brick[1][1], 32, 16, x, y)
	--initial position of shape wont matter since it will be moved to the center of the sprite
	local shape = shapes.newRectangleShape(0, 0, 32, 16)
	self:setShape(shape)
	self.gameType = "brick"
	self.health   = 10
	self.armor    = 1
	self.points   = 20

	self.overlap = {} --this doesnt have to be used by every single brick type
	self.checkOverlap = true
	self.suppress = false --suppressing a brick will usually prevent it doing special things like spawning bricks or dropping powerups

	self.patch = {shield = {},
				  cycle = {},
				  invisible = nil}
	self.shieldSprites = {}

	self.drawPriority = 0

	self.hitSound = "blockhit"
	self.deathSound = "blockbreak"

	self.brickType = "Brick"
	self.essential = true --determines whether or not this brick needs to be destroyed in order for the player to beat the level
end

function Brick:setPatches(patches)
	if patches.shield_up then self.patch.shield.up = true end
	if patches.shield_down then self.patch.shield.down = true end
	if patches.shield_left then self.patch.shield.left = true end
	if patches.shield_right then self.patch.shield.right = true end
	if patches.invisible then self.patch.invisible = true end
	if patches.antilaser then self.patch.antilaser = true end
	if patches.movement then
		local m = patches.movement
		self.patch.movement = {dir = m[1], spd = m[2], hit = m[3]}
	end
end

--called right before the game starts; after all the bricks are placed
function Brick:activate()
	local emplace = function(dir, i, j)
		if self.patch.shield[dir] then 
			local shield = Sprite:new("brick_spritesheet", rects.brick[i][j], 32, 16, self.x, self.y)
			local anistr = "Shield"..util.cap_first_letter(dir)
			function shield:playShieldAnimation()
				self:playAnimation(anistr)
			end
			self.shieldSprites[dir] = shield
		end
	end
	emplace("up", 7, 21)
	emplace("down", 6, 21)
	emplace("left", 7, 20)
	emplace("right", 6, 20)
	if self.patch.movement and not self.patch.movement.hit then
		self:activateMovement()
	end
	if self.patch.antilaser then
		self.lasershield = Sprite:new("patch_antilaser", rects.brick[1][1], 32, 16, self.x, self.y)
	end
end

brickSpeed = {slow = 125, medium = 250, fast = 500}
function Brick:activateMovement()
	local m = self.patch.movement
	local spd = brickSpeed[m.spd]
	if m.dir == "right" then
		self.vx = spd
	elseif m.dir == "left" then
		self.vx = -spd
	elseif m.dir == "up" then
		self.vy = -spd
	elseif m.dir == "down" then
		self.vy = spd
	end
	self.isMoving = true
	self.moveType = "constant"
end

function Brick:initSnapper()
	self.snapper = Sprite:new(nil, nil, 32, 16, self.x, self.y)
	self.snapper:playAnimation("SnapperMine", true)
end

--if you're extending this, call this function first then modify the copy
function Brick:clone() --something is still not right
	local c = util.copy(self)
	c.snapper = nil
	if self.isAnimating then
		c:playAnimation(self.currentAnistr, self.aniLoop)
	end
	c:setShape(shapes.newRectangleShape(0, 0, self.w, self.h))
	c.patch = {shield = {},
			   cycle = {},
			   invisible = nil,
			   antilaser = nil}
	c.overlap = {}
	return c
end

function Brick:kill()
	self.health = 0
end

function Brick:isDead()
	return self.health <= 0
end

function Brick:onDeath()
	playstate:incrementScore(self.points)

	if self.snapper then
		--explosion hitbox
		local e = Projectile:new("clear_pixel", nil, self.x, self.y, 0, 0, 0, "rectangle", 96, 48)
		e:setComponent("explosion")
		e.damage = 1000
		e.strength = 2
		game:emplace("callbacks", Callback:new(0.03, function() game:emplace("projectiles", e) end))
		--explosion smoke
		local i = math.random(0, 3)
		local smokeStr = "explosion_smoke"
		local p = Particle:new(smokeStr, {i*24, 0, 24, 24}, 50, 50, self.x, self.y, 0, 0, 0, 1)
		p.fadeRate = 750
		p.growthRate = 600
		p.growthAccel = -2000
		p.drawPriority = -1
		game:emplace("particles", p)
		--explosion sprite
		local anistr = "Explosion"
		local p = Particle:new("clear_pixel", nil, 96, 48, self.x, self.y, 0, 0, 0, 0.5)
		p:playAnimation(anistr)
		p.drawPriority = 1
		game:emplace("particles", p)
	end
end

function Brick.getHitSide(norm) --static
	local deg = math.atan2(norm.y, norm.x) * 180 / math.pi
	if deg >= -135 and deg < -45 then return "up"
	elseif deg >= -45 and deg < 45 then return "right"
	elseif deg >= 45 and deg < 135 then return "down"
	else return "left" end
end

function Brick:invisReveal()
	self.patch.invisible = nil
	for i = 1, 30 do
		local dx = math.random(0, 13)
		local dy = math.random(0, 5)
		local x = self.rect.x + dx
		local y = self.rect.y + dy
		local r = make_rect(x, y, 2, 2)
		local rad = math.random() * math.pi * 2
		local mag = math.random(100, 250)
		local p = Particle:new(self.imgstr, r, 3, 3, self.x+math.random(-25, 25), self.y+math.random(-12, 12), mag*math.cos(rad), mag*math.sin(rad), 0, 1.5)
		p.ay = 1000
		game:emplace("particles", p)
	end
	playSound("invisreveal")
end

function Brick:handlePatches(obj, norm)
	if self.patch.movement and self.patch.movement.hit and not self.isMoving then
		self:activateMovement()
	end

	if self.snapper then return true end
	if self.patch.invisible then
		self:invisReveal()
		if obj.strength < self.armor or obj.damage < self.health + 10 then
			return false
		end
	end
	local dir = Brick.getHitSide(norm)
	if self.patch.shield[dir] and obj.strength < 2 then
		self.shieldSprites[dir]:playShieldAnimation()
		return false
	end
	if self.patch.antilaser and obj.laser then
		return false
	end
	return true
end

function Brick:checkBallHit(ball)
	if self.intangible then return false end
	--check if their bounding boxes overlap first
	local box1 = {self.shape:bbox()}
	local box2 = {ball.shape:bbox()}
	if not util.bboxOverlap(box1, box2) then return false end
	--then use SAT for more precise check
	local check, mtvx, mtvy = ball.shape:collidesWith(self.shape)
	if not check then return false end
	--checks to see if ball has collided with the brick in the previous frame
	if self.checkOverlap then
		if self.overlap[ball] then
			self.overlap[ball] = 1
			return false
		end
		self.overlap[ball] = 1
	end
	--then check if the normal vector is actually valid
	local norm = {x = mtvx, y = mtvy}
	local dx, dy = self.x - ball.x, self.y - ball.y
	if util.angleBetween(norm.x, norm.y, dx, dy) < math.pi / 2 then return false end
	if not ball:validCollision(norm.x, norm.y) then return false end --redundant, but better safe than sorry
	return true, norm
end

function Brick:onBallHit(ball, norm)
	local check = false
	if self:handlePatches(ball, norm) then
		self:takeDamage(ball.damage, ball.strength)
		check = true
	end
	ball:onBrickHit(self, norm)
	return check
end

function Brick:checkMenacerHit(menacer)
	return self:checkBallHit(menacer)
end

function Brick:onMenacerHit(menacer, norm)
	menacer:onBrickHit(self, norm)
end

function Brick:checkProjectileHit(proj)
	if self.intangible then return false end
	local box1 = {self.shape:bbox()}
	local box2 = {proj.shape:bbox()}
	if not util.bboxOverlap(box1, box2) then return false end
	local check, mtvx, mtvy = proj.shape:collidesWith(self.shape)
	if not check then return false end
	if self.checkOverlap then
		if self.overlap[proj] then
			self.overlap[proj] = 1
			return false
		end
		self.overlap[proj] = 1
	end
	if proj.shapeType ~= "circle" then return true end
	local norm = {x = mtvx, y = mtvy}
	if proj.vx == 0 and proj.vy == 0 then return true, norm end
	local dx, dy = self.x - proj.x, self.y - proj.y
	if util.angleBetween(norm.x, norm.y, dx, dy) < math.pi / 2 then return false end
	if not proj:validCollision(norm.x, norm.y) then return false end
	return true, norm
end

function Brick:onProjectileHit(proj, norm)
	if self:handlePatches(proj, {x=proj.x-self.x, y=proj.y-self.y}) then
		self:takeDamage(proj.damage, proj.strength)
	end
	proj:onBrickHit(self, norm)
end

--damage subtracts from health only if strength is greater than or equal to armor
--NOTE: str means strength
function Brick:takeDamage(dmg, str)
	if self.snapper then self:kill() end
	if self.armor <= str then
		self.health = self.health - dmg
	end

	if self:isDead() then
		playSound(self.deathSound, false, self)
	else
		playSound(self.hitSound, false, self)
	end
end

function Brick:update(dt)
	--intangibility means that projectiles and balls would go through the brick instead of hitting it
	if self.intangible then
		if self.intangibleTimer then
			self.intangibleTimer = self.intangibleTimer - dt
			if self.intangibleTimer <= 0 then
				self.intangible = nil
				self.intangibleTimer = nil
			end
		end
	end
	--optional death timer
	if self.deathTimer then
		self.deathTimer = self.deathTimer - dt
		if self.deathTimer <= 0 then
			self:kill()
		end
	end
	--update shield sprites (if any)
	for _, v in pairs(self.shieldSprites) do
		v:update(dt)
	end
	--processes overlap objects
	for obj, state in pairs(self.overlap) do
		if state == 1 then
			self.overlap[obj] = 0
		else
			self.overlap[obj] = nil
		end
	end

	--movement
	if self.isMoving then
		if self.moveType == "target" then
			self.moveTimer = self.moveTimer - dt
			if self.moveTimer <= 0 then
				self.isMoving = false
				if self.intangibleWhileMoving then
					self.intangible = nil
					self.intangibleWhileMoving = nil
				end
				self:setPos(unpack(self.destination))
				self:updateShape() --makes sure the shape is absolutely aligned before checking
				self.vx, self.vy = 0, 0
				if self.x < window.lwallx 
				or self.x > window.rwallx 
				or self.y < window.ceiling
				or self.y > window.h then
					self:kill()
				end
				if self.moveCheck then
					self:moveCheck()
				end
			end
		else --moveType == "constant"
			-- local x, y = self:getPos()
			-- local w, h = self.w/2, self.h/2
			-- if x - w < window.lwallx then self.vx = math.abs(self.vx) end
			-- if x + w > window.rwallx then self.vx = -math.abs(self.vx) end
			-- if y - h < window.ceiling then self.vy = math.abs(self.vy) end
			-- if y + h > window.h then self.vy = -math.abs(self.vy) end
		end
	end

	Sprite.update(self, dt)

	--update snapper
	if self.snapper then
		self.snapper:update(dt)
		self.snapper:setPos(self:getPos())
	end

	for _, s in pairs(self.shieldSprites) do
		s:setPos(self:getPos())
	end

	if self.lasershield then self.lasershield:setPos(self:getPos()) end
end

function Brick:draw()
	if self.patch.invisible then return end

	--specialized drawing method that prevents graphical glitches (maybe obsolete)
	love.graphics.setColor(self.color.r, self.color.g, self.color.b, self.color.a)
	drawBrick(self.imgstr, self.rect, self.x, self.y, self.angle, self.w, self.h, nil, nil, nil, nil, self.drawFloor)
	--drawBrick(self.imgstr, self.rect, math.floor(self.x), math.floor(self.y) + 0.5, self.angle, self.w, self.h, nil, nil, nil, nil, false)
	--drawBrick(self.imgstr, self.rect, math.floor(self.x+0.5), math.floor(self.y+0.5) - 0.5, self.angle, self.w, self.h, nil, nil, nil, nil, false)

	--drawing order of shield sprites matters
	local s = self.shieldSprites
	if s.up then s.up:draw() end
	if s.down then s.down:draw() end
	if s.left then s.left:draw() end
	if s.right then s.right:draw() end

	if self.patch.antilaser then
		self.lasershield:draw()
	end

	--snappers
	if self.snapper then
		self.snapper:draw()
	end
end

function Brick:moveTo(x, y, time, funcstr)
	funcstr = funcstr or "null"
	self.isMoving = true
	self.moveType = "target"
	self.vx = (x - self.x) / time
	self.vy = (y - self.y) / time
	self.moveTimer = time
	self.destination = {x, y}
	self.moveCheck = Brick.moveCheck[funcstr]
end

function Brick:moveTo2(x, y, spd, funcstr)
	funcstr = funcstr or "null"
	self.isMoving = true
	self.moveType = "target"
	local dx, dy = x - self.x, y - self.y
	local dist = math.sqrt(dx*dx + dy*dy)
	self.vx = dx / dist * spd
	self.vy = dy / dist * spd
	self.moveTimer = dist / spd
	self.destination = {x, y}
	self.moveCheck = Brick.moveCheck[funcstr]
end

function Brick:inheritMovement(br)
	if not br.isMoving then return end
	self.isMoving = true
	self.moveType = br.moveType
	self.vx = br.vx
	self.vy = br.vy
	self.moveTimer = br.moveTimer
	self.destination = br.destination
	self.moveCheck = br.moveCheck
end

Brick.moveCheck = 
{
	null = function(self)
	end,
	die = function(self) --dies if it overlaps a brick
		local bucket = playstate:getBrickBucket(self)
		local box1 = {self.shape:bbox()}
		for br in pairs(bucket) do
			if br ~= self then
				local box2 = {br.shape:bbox()}
				if util.bboxOverlap(box1, box2) and not br.isMoving then
					self:kill()
					return
				end
			end
		end
	end,
	kill = function(self) --kills all bricks it overlaps
		local bucket = playstate:getBrickBucket(self)
		local box1 = {self.shape:bbox()}
		for br in pairs(bucket) do
			if br ~= self then
				local box2 = {br.shape:bbox()}
				if util.bboxOverlap(box1, box2) and not br.isMoving then
					br:kill()
					br.suppress = true
				end
			end
		end
	end,
	bounce = function(self) --bounces off of bricks
		local i, j = getGridPos(self:getPos())
		local border = i == 1 or i == 32 or j == 1 or j == 13
		local bucket = playstate:getBrickBucket(self)
		local box1 = {self.shape:bbox()}
		for br in pairs(bucket) do
			if br ~= self then
				local box2 = {br.shape:bbox()}
				if util.bboxOverlap(box1, box2) and not br.isMoving then
					candidates = {}
					for a = -1, 1 do
						for b = -1, 1 do
							if not (a == 0 and b == 0) and boundCheck(i + a, j + b) then
								table.insert(candidates, {i + a, j + b})
							end
						end
					end
					ii, jj = unpack(candidates[math.random(#candidates)])
					x, y = getGridPosInverse(ii, jj)
					self:moveTo2(x, y, 500, "bounce")
				end
			end
		end
	end
}

--Manages the movement of all bricks with movement patches.
--Attributes:
--	Brick has to bounce off of other bricks and walls
--  When one brick bounces off, all bricks that are aligned in the same axis should bounce at the same frame
--  How to deal with lag?
--Move all the bricks first
--Detect collisions and resolve them
--Maybe move the offsets to the end of the loop?
function signCmp(a, b)
	if a < 0 then return b < 0 else return b >= 0 end
end

function Brick.manageMovement()
	--divides the moving blocks into groups based on their velocities
	local mb = {} --abbreviation for moving bricks
	for _, br in pairs(game.bricks) do
		if not mb[br.vx] then mb[br.vx] = {} end
		if not mb[br.vx][br.vy] then mb[br.vx][br.vy] = {} end
		table.insert(mb[br.vx][br.vy], br)
	end
	--if we iterate through the non-moving bricks first... then maybe there wont be movement glitches?

	local movingBricks = {}
	for _, brick in pairs(game.bricks) do
		if brick.isMoving and brick.moveType == "constant" then
			table.insert(movingBricks, brick)
		end
	end
	for _, brick in pairs(movingBricks) do
		--local bucket = self:getBrickBucket(brick) --brick bucket may be out of sync, might need to update if i encounter any problems
		--for br, flag in pairs(bucket) do
		for _, br in pairs(game.bricks) do
			local dx, dy = br.x - brick.x, br.y - brick.y
			if br ~= brick and math.abs(dx) < 32 and math.abs(dy) < 16 then
				--local vkey, dkey = "vx", "dx"
				if 	   brick.vx > 0 then
					brick.x = br.x - 32
					brick.vx = -brick.vx
					Brick.cascade(brick, movingBricks, "x", -1)
				elseif brick.vx < 0 then
					brick.x = br.x + 32
					brick.vx = -brick.vx
					Brick.cascade(brick, movingBricks, "x", 1)
				elseif brick.vy > 0 then
					brick.y = br.y - 16
					brick.vy = -brick.vy
					Brick.cascade(brick, movingBricks, "y", -1)
				elseif brick.vy < 0 then
					brick.y = br.y + 16
					brick.vy = -brick.vy
					Brick.cascade(brick, movingBricks, "y", 1)
				end
			end
		end
		if 	   brick.x + 16 > window.rwallx then 
			brick.vx = -math.abs(brick.vx)
			brick.x = window.rwallx - 16
			Brick.cascade(brick, movingBricks, "x", -1)
		elseif brick.x - 16 < window.lwallx then 
			brick.vx = math.abs(brick.vx)
			brick.x = window.lwallx + 16
			Brick.cascade(brick, movingBricks, "x", 1)
		elseif brick.y + 8 > window.h then 
			brick.vy = -math.abs(brick.vy)
			brick.y = window.h - 8
			Brick.cascade(brick, movingBricks, "y", -1)
		elseif brick.y - 8 < window.ceiling then
			brick.vy = math.abs(brick.vy)
			brick.y = window.ceiling + 8
			Brick.cascade(brick, movingBricks, "y", 1)
		end
	end
end

function Brick.cascade(br1, movingBricks, axis, sign) --make it recursive?
	local val, vel, perp
	if axis == "x" then
		val, vel, perp = 32, "vx", "y"
	else
		val, vel, perp = 16, "vy", "x"
	end
	for _, br2 in pairs(movingBricks) do
		if br2[perp] == br1[perp] and br2[vel] * br1[vel] < 0 and util.deltaEqual(br1[axis] + val*sign, br2[axis], val/5) then
			br2[axis] = br1[axis] + val*sign
			br2[vel] = -br2[vel]
			Brick.cascade(br2, movingBricks, axis, sign)
		end
	end
end

--Null Brick dies after being placed
--used to resolve simultaneous brick placement
NullBrick = class("NullBrick", Brick)

function NullBrick:initialize(x, y)
	Brick.initialize(self, x, y)
	self.health = 0
	self.armor = 10
	self.points = 0
	self.brickType = "NullBrick"
end

function NullBrick:draw()
end

--Forbidden Brick prevents any brick from moving towards that location

ForbiddenBrick = class("ForbiddenBrick", Brick)

function ForbiddenBrick:initialize(x, y)
	Brick.initialize(self, x, y)
	self.health = 1000
	self.armor = 10
	self.points = 0
	self.brickType = "ForbiddenBrick"
end

--Patches will have no effect on this brick
function ForbiddenBrick:setPatches(patches)
end

function ForbiddenBrick:checkBallHit(ball)
	return false
end

function ForbiddenBrick:checkProjectileHit(proj)
	return false
end

function ForbiddenBrick:draw()
end

