Button = class("Button")

Button.defaultColor = {
	border = {0, 0, 0},
	idle = {bg = {255, 255, 255}, text = {0, 0, 0}},
	hovered = {bg = {255, 255, 0}, text = {0, 0, 0}},
	clicked = {bg = {255, 178, 102}, text = {0, 0, 0}}
}

Button.defaultColorToggle = {
	border = {0, 0, 0},
	idle = {on = {bg = {0, 255, 0}, text = {0, 0, 0}},
			off = {bg = {255, 0, 0}, text = {0, 0, 0}}},
	hovered = {bg = {255, 255, 0}, text = {0, 0, 0}},
	clicked = {bg = {255, 178, 102}, text = {0, 0, 0}}
}

local function containMouse(box)
	return mouse.x > box.x 
	   and mouse.x < box.x + box.w 
	   and mouse.y > box.y 
	   and mouse.y < box.y + box.h
end

function Button:initialize(x, y, w, h, options, callback)
	if type(x) == "table" then
		self.x, self.y, self.w, self.h = unpack(x)
		options, callback = y, w
	else
		self.x, self.y, self.w, self.h = x, y, w, h
	end
	self.callback = callback or util.nullfunc

	self.state = "idle"

	options = options or {} --maybe set default options table
	if options.imgstr then
		self.drawMode = "image"
		self.imgstr = options.imgstr
		self.rect = options.rect
	else
		self.drawMode = "plain"
		self.mode = options.mode or "standard" --standard or toggle
		self.text = options.text or ""
		self.font = options.font or font["Arcade30"]
		self.border = options.border or 2
		self.color = options.color or Button.defaultColor
		self.wrap = options.wrap or 1 --how many lines will the text wrap around
		if self.mode == "toggle" and self.color == Button.defaultColor then
			self.color = Button.defaultColorToggle
		end
	end
end

function Button:update(dt)
	local hit = false
	if containMouse(self) then
		if mouse.m1 then
			if mouse.m1 == 1 then
				self.state = "clicked"
			elseif self.state ~= "clicked" then
				self.state = "hovered"
			end
		elseif self.state == "clicked" then
			self.callback()
			hit = true
			self.state = "hovered"
		else
			self.state = "hovered"
		end
	elseif not (self.state == "clicked" and mouse.m1) then
		self.state = "idle"
	end
	if self.mode == "toggle" and hit then
		self.on = not self.on
	end
	return hit
end

function Button:draw()
	if self.drawMode == "plain" then
		if self.mode == "standard" then
			local color = self.color[self.state]
			love.graphics.setColor(0, 0, 0, 255)
			love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
			love.graphics.setColor(color.bg)
			love.graphics.rectangle("fill", self.x + self.border, self.y + self.border, self.w - self.border*2, self.h - self.border*2)
			love.graphics.setColor(color.text)
			love.graphics.setFont(self.font)
			love.graphics.printf(self.text, self.x, self.y + (self.h - self.font:getHeight() * self.wrap)/2, self.w, "center")
		elseif self.mode == "toggle" then
			local color
			if self.state == "idle" then
				color = self.color.idle[self.on and "on" or "off"]
			else
				color = self.color[self.state]
			end
			love.graphics.setColor(0, 0, 0, 255)
			love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
			love.graphics.setColor(color.bg)
			love.graphics.rectangle("fill", self.x + self.border, self.y + self.border, self.w - self.border*2, self.h - self.border*2)
			love.graphics.setColor(color.text)
			love.graphics.setFont(self.font)
			love.graphics.printf(self.text, self.x, self.y + (self.h - self.font:getHeight() * self.wrap)/2, self.w, "center")
		end
	end
end


TextBox = class("TextBox")

TextBox.defaultColor = {
	idle = {bg = {255, 255, 255}, text = {0, 0, 0}},
	selected = {bg = {255, 255, 0}, text = {0, 0, 0}}
}
TextBox.alwaysWhite = {
	idle = {bg = {255, 255, 255}, text = {0, 0, 0}},
	selected = {bg = {255, 255, 255}, text = {0, 0, 0}}
}

--callback is called whenever enter is pressed
function TextBox:initialize(x, y, w, h, font)
	self.x, self.y, self.w, self.h = x, y, w, h
	self.font = font
	self.text = ""
	self.color = TextBox.defaultColor
	self.selected = false
	self.fixedScale = nil
	self.off = {x = 0, y = 0}
	self.alwaysSelected = false --used for specific cases
end

function TextBox:update(dt)
	if mouse.m1 == 1 then
		self.selected = containMouse(self)
	end

	if self.selected or self.alwaysSelected then
		if keys.lastText then
			self.text = self.text .. keys.lastText
			return keys.lastText
		end
		if keys.backspace then
			if self.text:len() > 0 then
				self.text = self.text:sub(1, -2)
			end
			return "backspace"
		end
	end
end

function TextBox:draw()
	local color = self.color[self.selected and "selected" or "idle"]
	love.graphics.setColor(color.bg)
	love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
	love.graphics.setColor(color.text)
	love.graphics.setFont(self.font)
	love.graphics.print(self.text, self.x + self.off.x, self.y + self.off.y)
end


--NumBox = class("NumBox", TextBox)

NumBox = class("NumBox", TextBox)

--i could have used the tonumber function but whatever
NumBox.digits = util.generateLookup({"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"})

function NumBox:initialize(x, y, w, h, font)
	TextBox.initialize(self, x, y, w, h, font)
	self.text = "0"
	self.num = 0
end

function NumBox:update(dt) --returns true if the value is changed
	if mouse.m1 == 1 then
		self.selected = containMouse(self)
	end

	if self.selected then
		local oldNum = self.num
		if keys.lastText and NumBox.digits[keys.lastText] then
			local digit = keys.lastText
			if self.text == "0" then 
				self.text = "" 
			end
			self.text = self.text .. digit
		elseif keys["."] then
			if not self.period then 
				self.text = self.text .. "."
				self.period = true
			end
		elseif keys.backspace then
			if self.text:len() > 0 then
				if self.text:sub(-1) == "." then
					self.period = false
				end
				self.text = self.text:sub(1, -2)
			end
			if self.text == "" then
				self.text = "0"
			end
		end
		self.num = tonumber(self.text)
		return oldNum ~= self.num
	end
	return false
end

--assuming the arguments are valid
function NumBox:setText(text)
	self.text = text
	self.num = tonumber(text)
	self:updatePeriod()
end 

function NumBox:setNumber(n)
	self.text = tostring(n)
	self.num = n
	self:updatePeriod()
end

function NumBox:updatePeriod()
	self.period = false
	for i = 1, #self.text do
		if self.text:sub(i,i) == "." then
			self.period = true
		end
	end
end

CheckBox = class("CheckBox")

CheckBox.defaultBackground = {0, 0, 0, 0}

function CheckBox:initialize(x, y, scale, state, background, manual)
	self.x = x
	self.y = y
	scale = scale or 1
	self.w = 16 * scale
	self.h = 16 * scale
	self.rect = make_rect(0, 0, 16, 16)
	self.state = state == true --should be boolean
	self.background = background or CheckBox.defaultBackground
	self.manual = manual == true --instead of toggling, you can make the checkbox return whether or not it was pressed
end

function CheckBox:update(dt)
	if containMouse(self) and mouse.m1 == 1 then
		if self.manual then return true end
		self.state = not self.state
	end
	if self.manual then return false end
end

function CheckBox:draw()
	love.graphics.setColor(unpack(self.background))
	love.graphics.rectangle("fill", self.x, self.y, self.w, self.h)
	love.graphics.setColor(255, 255, 255, 255)
	self.rect[1] = self.state and 0 or 16
	draw("checkbox", self.rect, self.x + self.w/2, self.y + self.h/2, 0, self.w, self.h)
end

Slider = class("Slider")

function Slider:initialize(x, y, w, h, min, max, initial, tics)
	self.x = x
	self.y = y
	self.w = w
	self.h = h
	self.min = min
	self.max = max
	self.barw = 10
	self.value = initial
	self.tics = tics or 10
end

function Slider:valueToX(value)
	value = value or self.value
	return self.x + self.w * (self.value - self.min) / (self.max - self.min)
end

function Slider:xToValue(x)
	x = x or self.x
	local dx = x - self.x
	return self.min + (dx / self.w) * (self.max - self.min)
end

function Slider:update(dt)
	if mouse.m1 then
		local mx, my = mouse.x, mouse.y
		if self.drag or mx > self.x and mx < self.x + self.w and my > self.y and my < self.y + self.h then
			self.drag = true
			self.value = self:xToValue(mx)
			self.value = math.max(self.min, math.min(self.max, math.floor(self.value + 0.5)))
			return self.value
		end
	else
		self.drag = nil
	end
end

function Slider:draw()
	love.graphics.setLineWidth(2)
	love.graphics.setLineStyle("rough")
	love.graphics.setColor(0, 0, 0, 255)

	local x, y, w, h = self.x, self.y, self.w, self.h
	love.graphics.line(x, y, x, y+h)
	love.graphics.line(x, y+h/2, x+w, y+h/2)
	love.graphics.line(x+w, y, x+w, y+h)

	love.graphics.setLineWidth(1)
	for i = 0, self.tics do
		local dx = i*w/self.tics
		love.graphics.line(x+dx, y+h/4, x+dx, y+3*h/4)
	end

	love.graphics.setColor(255, 255, 255, 255)
	local min, max, barw = self.min, self.max, self.barw
	local barx = self:valueToX()
	love.graphics.rectangle("fill", barx - barw/2, y, barw, h)
end