LevelSelectState = class("LevelSelectState")

local function drawWidget(widget)
	local border = 2
	love.graphics.setColor(0, 0, 0, 255)
	love.graphics.rectangle("fill", widget.x, widget.y, widget.w, widget.h)
	love.graphics.setColor(150, 150, 150, 255)
	love.graphics.rectangle("fill", widget.x + border, widget.y + border, widget.w - border*2, widget.h - border*2)
end

local function unpackDim(box)
	return box.x, box.y, box.w, box.h
end

function LevelSelectState.generatePreview(file)
	local chunk = nil
	if file:sub(1, 1) == "_" then
		chunk = love.filesystem.load("default_levels/"..file)
	else
		chunk = love.filesystem.load("levels/"..file)
	end
	if not chunk then return "error" end
	local data = chunk()
	if not data then return "error" end

	local canvas = love.graphics.newCanvas(208, 256)
	love.graphics.setCanvas(canvas)
	for _, v in pairs(data.bricks) do
		local i, j, id, patch = unpack(v)
		local x, y = j*16-8, i*8-4
		local brickData = EditorState.brickDataLookup[id]
		brickData:draw(x, y, 16, 8)
		for k, v in pairs(patch) do
			EditorState.drawPatch(k, v, x, y, 16, 8)
		end
	end
	love.graphics.setCanvas()
	return canvas
end

function LevelSelectState.readPlaylist(file)
	local lines = {}
	local iter
	if file:sub(1, 1) == "_" then
		iter = love.filesystem.lines("default_playlists/"..file)
	else
		iter = love.filesystem.lines("playlists/"..file)
	end
	for line in iter do
		table.insert(lines, line)
	end
	return lines
end

--self.files is a list of pairs containing the filename and preview canvas
function LevelSelectState:loadFileList()
	love.filesystem.createDirectory("levels") --creates a folder in the save directory if missing
	love.filesystem.createDirectory("playlists")

	local default, user
	if self.mode == "playlist" then
		default = love.filesystem.getDirectoryItems("default_playlists")
		user = love.filesystem.getDirectoryItems("playlists")
	else
		default = love.filesystem.getDirectoryItems("default_levels")
		user = love.filesystem.getDirectoryItems("levels")
		if self.mode == "editor_save" then
			default = {}
		end
	end
	
	self.files = {}
	local sz = 0
	
	for i, f in ipairs(default) do
		self.files[i] = {f}
		sz = sz + 1
	end
	for i, f in ipairs(user) do
		self.files[i+sz] = {f}
	end
	util.remove_if(self.files, function(file) return file == ".DS_Store" end)
	self.numfiles = #self.files
end

--returns the index of the file and the index of the closest match
function LevelSelectState:findFile(name)
	local start = (self.mode == "editor_save") and 2 or 1
	local index, close_index = 0, 0
	--assuming the files are listed in alphabetical order
	for i = start, self.numfiles do
		local file = self.files[i]
		if file[1] == name then
			if index == 0 then
				index = i
			end
		end
		if string.sub(file[1], 1, #name) == name then
			if close_index == 0 then
				close_index = i
			end
		end
	end
	return index, close_index
end

function LevelSelectState:initialize(mode, startFile)
	--"level_select", "editor_save", "editor_load", "playlist"
	self.mode = mode or "level_select"
	self.newFileString = "===NEW FILE==="

	self:loadFileList()
	if self.mode == "editor_save" then
		table.insert(self.files, 1, {self.newFileString, "error"})
		self.numfiles = self.numfiles + 1
	end

	local lwidget, rwidget, dwidget
	lwidget = {
		x = 50, 
		y = 50, 
		w = window.w/2 - 55, 
		h = window.h - 150
	}
	lwidget.textbox = {
		x = lwidget.x + 20,
		y = lwidget.y + 20,
		w = lwidget.w - 40 - 20,
		h = lwidget.h - 40 + 4, --now it is tall enough for exactly 23 rows of 18 pixels tall
		index = 1,
		offset = 0
	}
	lwidget.scrollbar = {
		x = lwidget.x + 300 + 10,
		y = lwidget.y + 20,
		w = 15,
		h = lwidget.h - 40 + 4,
	}
	--these methods will also be used by rwidget for displaying a playlist file
	lwidget.textbox.update = function(tbox, dt)
		local numfiles = math.max(1, self.numfiles)
		if keys.up or keys.down then
			if keys.up then
				tbox.index = math.max(1, tbox.index - 1)
			else
				tbox.index = math.min(numfiles, tbox.index + 1)
			end
			tbox.offset = math.max(math.min(tbox.offset, tbox.index - 1), tbox.index - 23)
			dwidget.textbox.text = self.files[tbox.index][1]
			tbox.transparent = false
		end
		if mouse.m1 == 1 then
			local mx, my = mouse.x, mouse.y
			if mx > tbox.x and mx < tbox.x + tbox.w and my > tbox.y and my < tbox.y + tbox.h then
				local dy = my - tbox.y
				local row = math.floor(dy / 18) + 1
				tbox.index = row + tbox.offset
				tbox.index = math.min(numfiles, tbox.index)
				local file = self.files[tbox.index]
				dwidget.textbox.text = file[1]
				tbox.transparent = false
				if self.mode == "playlist" and not file[2] then
					file[2] = LevelSelectState.readPlaylist(file[1])
				end
			end
		end
		
	end
	lwidget.textbox.draw = function(tbox)
		love.graphics.setColor(200, 200, 200, 255)
		love.graphics.rectangle("fill", unpackDim(tbox))
		love.graphics.setScissor(unpackDim(tbox))
		love.graphics.setColor(255, 255, 0, tbox.transparent and 128 or 255)
		love.graphics.rectangle("fill", tbox.x, tbox.y + (tbox.index-1-tbox.offset)*18, tbox.w, 18)
		love.graphics.setFont(font["Munro20"])
		for i, file in ipairs(self.files) do
			if file[1]:sub(1, 1) == "_" then
				love.graphics.setColor(0, 0, 255, 255)
			else
				love.graphics.setColor(0, 0, 0, 255)
			end
			love.graphics.print(file[1], tbox.x + 2, tbox.y + (i-1-tbox.offset)*18 - 2)
		end
		love.graphics.setScissor()
	end
	lwidget.scrollbar.update = function(sc, dt)
		local numfiles = math.max(1, self.numfiles)
		sc.bar = {}
		local bar = sc.bar
		bar.x = sc.x
		bar.w = sc.w
		bar.h = sc.h * math.min(23 / numfiles, 1)
		bar.y = sc.y + lwidget.textbox.offset * sc.h / numfiles
		if mouse.m1 == 1 then
			local mx, my = mouse.x, mouse.y
			if mx > bar.x and mx < bar.x + bar.w and my > bar.y and my < bar.y + bar.h then
				sc.drag = {my, lwidget.textbox.offset}
			end
		elseif mouse.m1 == 2 then
			if sc.drag then
				local offset = lwidget.textbox.offset
				local dy = mouse.y - sc.drag[1]
				local doff = math.floor((numfiles) * (dy / sc.h))
				offset = math.min(math.max(sc.drag[2] + doff, 0), math.max(numfiles - 23, 0))
				lwidget.textbox.offset = offset
			end
		else
			sc.drag = nil
		end
		if mouse.scrollup then
			lwidget.textbox.offset = lwidget.textbox.offset - 1
			lwidget.textbox.offset = math.min(math.max(lwidget.textbox.offset, 0), math.max(numfiles - 23, 0))
		end
		if mouse.scrolldown then
			lwidget.textbox.offset = lwidget.textbox.offset + 1
			lwidget.textbox.offset = math.min(math.max(lwidget.textbox.offset, 0), math.max(numfiles - 23, 0))
		end
	end
	lwidget.scrollbar:update(0) --have to call this function to make sure bar is initialized
	lwidget.scrollbar.draw = function(sc)
		love.graphics.setColor(200, 200, 200, 255)
		love.graphics.rectangle("fill", sc.x, sc.y, sc.w, sc.h)
		love.graphics.setColor(110, 110, 110, 255)
		love.graphics.rectangle("fill", sc.bar.x, sc.bar.y, sc.bar.w, sc.bar.h)
	end
	lwidget.update = function(widget, dt)
		widget.textbox:update(dt)
		widget.scrollbar:update(dt)
	end
	lwidget.draw = function(widget)
		drawWidget(widget)
		widget.textbox:draw()
		widget.scrollbar:draw()
	end

	rwidget = {
		x = window.w/2 + 5, 
		y = 50, 
		w = window.w/2 - 55, 
		h = window.h - 150,
	}
	if self.mode ~= "playlist" then
		local pw, ph = 208, 256
		rwidget.preview = {
			x = math.floor(rwidget.x + rwidget.w/2 - pw/2),
			y = math.floor(rwidget.y + 40),
			w = pw,
			h = ph
		}
		rwidget.draw = function(widget)
			local px, py, pw, ph = unpackDim(widget.preview)
			drawWidget(widget)
			love.graphics.setColor(0, 0, 0, 255)
			love.graphics.rectangle("fill", px, py, pw, ph)
			love.graphics.setColor(255, 255, 255, 255)
			draw("border", nil, px+pw/2, py+ph/2-4)
			local index = lwidget.textbox.index
			if index > 0 and index <= self.numfiles then
				local file = self.files[lwidget.textbox.index]
				if not file[2] then
					file[2] = LevelSelectState.generatePreview(file[1])
				end
				if type(file[2]) == "string" then
					love.graphics.setColor(255, 255, 255, 255)
					love.graphics.setFont(font["Munro30"])
					love.graphics.printf("No Preview Available", px, py + 20, pw, "center")
				else
					love.graphics.draw(file[2], px, py)
				end
				if file[1]:sub(1, 1) == "_" then
					love.graphics.setColor(0, 0, 255, 255)
					love.graphics.setFont(font["Munro20"])
					love.graphics.print("Default Level", px, py + ph)
				end
			end
		end
	else
		--I can borrow some elements from lwidget to copy a scroll bar
		rwidget.textbox = {
			x = rwidget.x + 20,
			y = rwidget.y + 20,
			w = rwidget.w - 40 - 20,
			h = rwidget.h - 40 + 4, --now it is tall enough for exactly 23 rows of 18 pixels tall
			index = 1,
			offset = 0
		}
		rwidget.textbox.draw = function(tbox)
			love.graphics.setColor(200, 200, 200, 255)
			love.graphics.rectangle("fill", unpackDim(tbox))
			love.graphics.setScissor(unpackDim(tbox))
			-- love.graphics.setColor(255, 255, 0, tbox.transparent and 128 or 255)
			-- love.graphics.rectangle("fill", tbox.x, tbox.y + (tbox.index-1-tbox.offset)*18, tbox.w, 18)
			love.graphics.setColor(0, 0, 0, 255)
			love.graphics.setFont(font["Munro20"])
			local index = lwidget.textbox.index
			if index > 0 and index <= self.numfiles then
				local lines = self.files[index][2]
				for i, line in ipairs(lines) do
					love.graphics.print(line, tbox.x + 2, tbox.y + (i-1-tbox.offset)*18 - 2)
				end
			end
			love.graphics.setScissor()
		end
		rwidget.draw = function(widget)
			drawWidget(widget)
			widget.textbox:draw()
		end
		--initialize first read
		if self.numfiles > 0 then
			local file = self.files[1]
			file[2] = LevelSelectState.readPlaylist(file[1])
		end
	end

	dwidget = {
		x = 50,
		y = window.h - 90,
		w = window.w - 100,
		h = 60
	}
	-- dwidget.textbox = {
	-- 	x = dwidget.x + 20,
	-- 	y = dwidget.y + 20,
	-- 	w = 300,
	-- 	h = 20,
	-- }
	dwidget.textbox = TextBox:new(dwidget.x + 20, dwidget.y + 10, 300, 18, font["Munro20"])
	dwidget.textbox.off = {x = 2, y = -2}
	dwidget.textbox.color = TextBox.alwaysWhite
	dwidget.textbox.alwaysSelected = true
	if self.numfiles > 0 then dwidget.textbox.text = self.files[1][1] end
	dwidget.playbutton = Button:new(dwidget.x + 360, dwidget.y + 10, 100, 40, {text = "Play", font = font["Arcade30"]})
	dwidget.savebutton = Button:new(dwidget.x + 360, dwidget.y + 10, 100, 40, {text = "Save", font = font["Arcade30"]})
	dwidget.loadbutton = Button:new(dwidget.x + 360, dwidget.y + 10, 100, 40, {text = "Load", font = font["Arcade30"]})
	dwidget.backbutton = Button:new(dwidget.x + 580, dwidget.y + 10, 100, 40, {text = "Back", font = font["Arcade30"]})
	dwidget.setMessage = function(widget, message, color)
		widget.message = message
		widget.messageColor = color or {0, 0, 0}
		widget.messageTimer = 3
	end
	dwidget.update = function(widget, dt)
		if widget.message then
			widget.messageTimer = widget.messageTimer - dt
			if widget.messageTimer <= 0 then
				widget.message = nil
			end
		end

		local v = widget.textbox:update(dt)
		if v or self.override then
			local tbox = lwidget.textbox
			local text = widget.textbox.text
			local index, close_index = self:findFile(text)
			if self.mode == "editor_save" then
				if index == 0 then index = 1 end
				tbox.index = index
			else
				tbox.index = close_index
				tbox.transparent = (index ~= close_index)
			end
			if tbox.index ~= 0 then
				tbox.offset = math.max(math.min(tbox.offset, tbox.index - 1), tbox.index - 23)
			end
		end
		if keys.tab then
			if lwidget.textbox.transparent then
				lwidget.textbox.transparent = false
				widget.textbox.text = self.files[lwidget.textbox.index][1]
			end
		end
		if self.mode == "editor_save" then
			if widget.textbox.text == self.newFileString then
				widget.textbox.text = ""
			end
			if widget.savebutton:update(dt) then
				local text = widget.textbox.text
				local tbox = lwidget.textbox
				if text == "" then
					widget:setMessage("ERROR: filename cannot be blank", {255, 0, 0})
				elseif text:sub(1, 1) == "_" then
					widget:setMessage("ERROR: filename cannot start with '_'", {255, 0, 0})
				else
					editorstate:saveBricks(text)
					if tbox.index == 1 then --new file is being created
						local index = 2
						while index <= self.numfiles and text > self.files[index][1] do
							index = index + 1
						end
						table.insert(self.files, index, {text, self.generatePreview(text)})
						tbox.index = index
						tbox.offset = math.max(math.min(tbox.offset, tbox.index - 1), tbox.index - 23)
					else --update on old file
						self.files[tbox.index][2] = self.generatePreview(text)
					end
					widget:setMessage("File saved!", {0, 255, 0})
				end
			end
		elseif self.mode == "editor_load" then
			if widget.loadbutton:update(dt) then
				if self:findFile(widget.textbox.text) == 0 then
					widget:setMessage("ERROR: file not found", {255, 0, 0})
				else
					editorstate:loadBricks(widget.textbox.text)
					game:pop()
				end
			end
		elseif self.mode == "playlist" then
			if widget.playbutton:update(dt) then
				local index = lwidget.textbox.index
				if index > 0 and index <= self.numfiles then
					local q = Queue:new(self.files[index][2])
					game:push(PlayState:new("play", q))
				end
			end
		else
			if widget.playbutton:update(dt) then
				if self:findFile(widget.textbox.text) == 0 then
					widget:setMessage("ERROR: file not found", {255, 0, 0})
				else
					game:push(PlayState:new("play", widget.textbox.text))
				end
			end
		end
		if widget.backbutton:update(dt) then
			game:pop();
		end
	end
	dwidget.draw = function(widget)
		drawWidget(widget)
		widget.textbox:draw()
		if widget.message then
			love.graphics.setFont(font["Munro20"])
			love.graphics.setColor(unpack(widget.messageColor))
			love.graphics.print(widget.message, widget.x + 20, widget.y + 30)
		end
		if self.mode == "editor_save" then
			widget.savebutton:draw()
		elseif self.mode == "editor_load" then
			widget.loadbutton:draw()
		else
			widget.playbutton:draw()
		end
		widget.backbutton:draw()
	end

	self.lwidget = lwidget
	self.rwidget = rwidget
	self.dwidget = dwidget

	if startFile then
		self.override = true
		dwidget.textbox.text = startFile
		dwidget:update(0)
		self.override = false
	end

end

function LevelSelectState:update(dt)
	if keys.escape then
		game:pop()
		return
	end

	self.lwidget:update(dt)
	self.dwidget:update(dt)

end

function LevelSelectState:draw()
	love.graphics.setColor(100, 100, 100, 255)
	love.graphics.rectangle("fill", 0, 0, window.w, window.h)

	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.setFont(font["Arcade50"])

	local text = "Level Select"
	if self.mode == "editor_save" then 
		text = "Save Bricks"
	elseif self.mode == "editor_load" then
		text = "Load Bricks"
	elseif self.mode == "playlist" then
		text = "Playlists"
	end
	love.graphics.printf(text, 0, 0, window.w, "center")

	self.lwidget:draw()
	self.rwidget:draw()
	self.dwidget:draw()
end